"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Handles attaching session
 */
const Session = require("msgpack5rpc");
const decode_1 = require("../utils/decode");
const generateWrappers_1 = require("./helpers/generateWrappers");
const types_1 = require("./helpers/types");
const Neovim_1 = require("./Neovim");
class NeovimClient extends Neovim_1.Neovim {
    constructor(options = {}) {
        const session = options.session || new Session([]);
        const { logger } = options;
        // Neovim has no `data` or `metadata`
        super({
            logger,
            session,
        });
        this.requestQueue = [];
        this._sessionAttached = false;
        this.handleRequest = this.handleRequest.bind(this);
        this.handleNotification = this.handleNotification.bind(this);
    }
    attachSession({ reader, writer }) {
        this._session.attach(writer, reader);
        this._sessionAttached = true;
        this.startSession();
    }
    get isApiReady() {
        return this._sessionAttached && typeof this._channel_id !== 'undefined';
    }
    handleRequest(method, args, resp, ...restArgs) {
        this.logger.info('handleRequest: ', method);
        // If neovim API is not generated yet and we are not handle a 'specs' request
        // then queue up requests
        //
        // Otherwise emit as normal
        if (!this.isApiReady && method !== 'specs') {
            this.requestQueue.push({
                type: 'request',
                args: [method, args, resp, ...restArgs],
            });
        }
        else {
            this.emit('request', decode_1.decode(method), decode_1.decode(args), resp);
        }
    }
    handleNotification(method, args, ...restArgs) {
        this.logger.info('handleNotification: ', method);
        // If neovim API is not generated yet then queue up requests
        //
        // Otherwise emit as normal
        if (!this.isApiReady) {
            this.requestQueue.push({
                type: 'notification',
                args: [method, args, ...restArgs],
            });
        }
        else {
            this.emit('notification', decode_1.decode(method), decode_1.decode(args));
        }
    }
    // Listen and setup handlers for session
    startSession() {
        if (!this._sessionAttached) {
            throw new Error('Not attached to input/output');
        }
        this._session.on('request', this.handleRequest);
        this._session.on('notification', this.handleNotification);
        this._session.on('detach', () => {
            // this.logger.debug('detached');
            this.emit('disconnect');
            this._session.removeAllListeners('request');
            this._session.removeAllListeners('notification');
            this._session.removeAllListeners('detach');
        });
        this._isReady = this.generateApi();
    }
    requestApi() {
        return new Promise((resolve, reject) => {
            this._session.request('nvim_get_api_info', [], (err, res) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve(res);
                }
            });
        });
    }
    // Request API from neovim and augment this current class to add these APIs
    generateApi() {
        return __awaiter(this, void 0, void 0, function* () {
            let results;
            try {
                results = yield this.requestApi();
            }
            catch (err) {
                this.logger.error('Could not get vim api results');
                this.logger.error(err);
            }
            if (results) {
                try {
                    const [channelId, encodedMetadata] = results;
                    const metadata = decode_1.decode(encodedMetadata);
                    const extTypes = [];
                    const types = {};
                    const prefixMap = {};
                    // this.logger.debug(`$$$: ${metadata}`);
                    Object.keys(metadata.types).forEach(name => {
                        let ExtType;
                        // Generate a constructor function for each type in metadata.types
                        if (typeof types_1.TYPES[name] === 'undefined') {
                            this.logger.warn(`Class not found for ${name}`);
                        }
                        else {
                            ExtType = types_1.TYPES[name];
                        }
                        const metaDataForType = metadata.types[name];
                        // Collect the type information necessary for msgpack5 deserialization
                        // when it encounters the corresponding ext code.
                        extTypes.push({
                            constructor: ExtType,
                            code: metaDataForType.id,
                            decode: data => new ExtType({
                                session: this._session,
                                data,
                                metadata: metaDataForType,
                                logger: this.logger,
                            }),
                            encode: obj => obj._data,
                        });
                        prefixMap[metaDataForType.prefix] = name;
                        types[name] = {
                            constructor: ExtType,
                            prefix: metaDataForType.prefix,
                        };
                        Neovim_1.Neovim.prototype[name] = ExtType;
                    });
                    generateWrappers_1.generateWrappers(Neovim_1.Neovim, types, prefixMap, metadata);
                    this._channel_id = channelId;
                    this._session.addTypes(extTypes);
                    // register the non-queueing handlers
                    // dequeue any pending RPCs
                    this.requestQueue.forEach(pending => {
                        this.emit(pending.type, ...pending.args);
                    });
                    this.requestQueue = [];
                    return true;
                }
                catch (err) {
                    this.logger.error(`Could not dynamically generate neovim API: ${err}`, {
                        error: err,
                    });
                    this.logger.error(err.stack);
                    return null;
                }
            }
            return null;
        });
    }
}
exports.NeovimClient = NeovimClient;
